"""
couchdiscover.kube
~~~~~~~~~~~~~~~~~~

This module contains constants and classes that abstract the details of
working with the kubernetes api as higher level objects.

:copyright: (c) 2016 by Joe Black.
:license: Apache2.
"""

import base64

import pykube

from . import config, util
from .exceptions import InvalidKubeHostnameError


class KubeHostname:
    """Represents a kubernetes hostname.

    A Kubernetes hostname where individual parts can be manipulated and the
    hostname is updated.  str(host) will return the hostname as a string.
    """

    def __init__(self, fqdn):
        self._init_from_fqdn(fqdn)

    def _verify_fqdn(self, fqdn):
        if len(fqdn.split('.')) < 5:
            raise InvalidKubeHostnameError(fqdn)

    @property
    def fqdn(self):
        """Returns a calculated FQDN for the hostname."""
        return self._join_fqdn(
            self.node, self.service, self.namespace, self.domain)

    @fqdn.setter
    def fqdn(self, fqdn):
        """"Allows a new kube hostname object to be set by manipulating the
        FQDN.
        """
        self._init_from_fqdn(fqdn)

    @property
    def node(self):
        """Retrieves the first part of the kubernetes hostname.
        Example: `couchdb-0`
        """
        return self._join_node(self.petset, self.index)

    @node.setter
    def node(self, node):
        """Allows the first part of the kubernetes hostname to be set and a
        new hostname object recalculated.
        """
        self.petset, self.index = self._split_node(node)

    @property
    def is_master(self):
        """Determines whether the index is 0 for when the master of a PetSet
        needs to be determined."""
        return self.index == 0

    def __repr__(self):
        class_name = type(self).__name__
        return '{}(\'{}\')'.format(class_name, self.fqdn)

    def __str__(self):
        return self.fqdn

    @staticmethod
    def _split_fqdn(fqdn):
        node, service, namespace, _, domain = fqdn.split('.', 4)
        return node, service, namespace, domain

    @staticmethod
    def _split_node(node):
        petset, index = node.split('-')
        index = int(index)
        return petset, index

    def _init_from_fqdn(self, fqdn):
        node, service, namespace, domain = self._split_fqdn(fqdn)
        petset, index = self._split_node(node)
        self.service = service
        self.namespace = namespace
        self.domain = domain
        self.petset = petset
        self.index = index

    @staticmethod
    def _join_fqdn(node, service, namespace, domain):
        return '.'.join((node, service, namespace, 'svc', domain))

    @staticmethod
    def _join_node(petset, index):
        return '{}-{}'.format(petset, str(index))

    def clone(self, master=False, index=None):
        """Clone's a copy of the current KubeHostname object.

        Options include requesting the KubeHostname for the master or by
        passing an index.
        """
        new = type(self)(self.fqdn)
        if index:
            new.index = index
        if master:
            new.index = 0
        return new


class KubeAPIClient:
    """Contains the lower level functions for manipulating and retrieving
    objects from the kubernetes api.
    """
    def __init__(self, env=None, namespace=None):
        self.env = env
        self.namespace = namespace
        self.api = self._get_api()

    def _get_api(self):
        if self.env == 'dev':
            api = pykube.http.HTTPClient(
                pykube.KubeConfig.from_file(config.DEV_KUBECONFIG_PATH))
        else:
            api = pykube.http.HTTPClient(
                pykube.KubeConfig.from_service_account())
        return api

    def _get_api_object(self, resource, name=None, selector=None,
                        namespace=None):
        if not namespace:
            namespace = self.namespace
        if not issubclass(resource, pykube.objects.APIObject):
            raise pykube.PyKubeError('No object by type: %s', resource)
        req = pykube.query.Query(
            self.api, resource, namespace=namespace)
        if name:
            req = req.get_by_name(name)
        if selector:
            req = req.filter(selector=selector)
        if req.exists:
            return req.obj

    def get_pod(self, name=None, selector=None, namespace=None):
        """Get's pod by name or/or selector."""
        return self._get_api_object(pykube.Pod, name, selector, namespace)

    def get_service(self, name=None, selector=None, namespace=None):
        """Get's service by name or/or selector."""
        return self._get_api_object(pykube.Service, name, selector, namespace)

    def get_endpoint(self, name=None, selector=None, namespace=None):
        """Get's endpoint by name or/or selector."""
        return self._get_api_object(pykube.Endpoint, name, selector, namespace)

    def get_petset(self, name=None, selector=None, namespace=None):
        """Get's petset by name or/or selector."""
        return self._get_api_object(pykube.PetSet, name, selector, namespace)

    def get_secret(self, name=None, key=None, selector=None, namespace=None):
        """Get's secret by name or/or selector."""
        sec = self._get_api_object(pykube.Secret, name, selector, namespace)
        if key:
            sec = self._get_key_decoded(sec, key)
        return sec

    def get_configmap(self, name=None, key=None, selector=None,
                      namespace=None):
        """Get's configmap by name or/or selector."""
        cm = self._get_api_object(pykube.ConfigMap, name, selector, namespace)
        if key:
            cm = cm['data'].get(key)
        return cm

    @staticmethod
    def _get_key_decoded(obj, key):
        val = obj['data'].get(key)
        if val:
            obj = base64.b64decode(val).decode()
        return obj

    @staticmethod
    def _get_container(petset=None, container=None):
        if petset:
            pod = petset['spec']['template']
        if container:
            for cont in pod['spec']['containers']:
                if cont['name'] == container:
                    return cont
        else:
            return pod

    @staticmethod
    def _key_container_env(container):
        return {item['name']: item for item in container['env']}

    def _lookup_env_value(self, env):
        if env.get('value'):
            return env.get('value')
        elif env.get('valueFrom'):
            v_from = env.get('valueFrom')
            if v_from.get('fieldRef'):
                # don't implement downward api here, unnecessary
                return ''
            elif v_from.get('secretKeyRef'):
                ref = v_from.get('secretKeyRef')
                return self.get_secret(name=ref['name'], key=ref['key'])
            elif v_from.get('configMapKeyRef'):
                ref = v_from.get('configMapKeyRef')
                return self.get_configmap(name=ref['name'], key=ref['key'])

    def get_environment(self, petset, container):
        """Get's the environment for a container of a petset.

        Environment returned is a dictionary key'd by environment variable
        name, and who's value has been resolved in the case of externally
        referenced configmaps and secrets.
        """
        petset = self.get_petset(petset)
        cont = self._get_container(petset, container)
        env = self._key_container_env(cont)
        return {k: self._lookup_env_value(v) for k, v in env.items()}


class KubeInterface(util.ReprMixin):
    """This class exposes the information we need from kubernetes as lazy
    evaluated properties, caching provided by pykube.
    """
    _public_attrs = ('hosts', 'ports', 'creds', 'cluster_size')

    def __init__(self, host, env=None):
        self._host = host
        self.api = KubeAPIClient(env=env, namespace=self._host.namespace)

    def _fqdn_from_node(self, node):
        host = self._host.clone()
        host.node = node
        return str(host)

    @property
    def hosts(self):
        """Returns a tuple of full fqdn's for all nodes in the CouchDB
        petset.
        """
        service = self._host.service
        endp = self.api.get_endpoint(service)
        hosts = [self._fqdn_from_node(address['hostname'])
                 for address in endp['subsets'][0]['addresses']]
        return tuple(sorted(hosts))

    @property
    def ports(self):
        """Returns a tuple of ports for the CouchDB petset."""
        service = self._host.service
        endp = self.api.get_endpoint(service)
        ports = [port['port'] for port in endp['subsets'][0]['ports']]
        return tuple(sorted(ports))

    @property
    def creds(self):
        """Returns a tuple of user/pass for the CouchDB petset."""
        petset = self._host.petset
        env = self.api.get_environment(petset, petset)
        user = env.get('COUCHDB_ADMIN_USER', config.DEFAULT_CREDS[0])
        password = env.get('COUCHDB_ADMIN_PASS', config.DEFAULT_CREDS[1])
        return (user, password)

    @property
    def cluster_size(self):
        """Returns the expected cluster size by qerying the
        `COUCHDB_CLUSTER_SIZE` environment variable, and falling back to the
        number of replicas in the petset.  For most purposes, you can rely on
        default behavior here.
        """
        petset = self._host.petset
        pobj = self.api.get_petset(petset)
        env = self.api.get_environment(petset, petset)
        size = pobj['spec']['replicas']
        return env.get('COUCHDB_CLUSTER_SIZE', size)



"""
couchdiscover.manage
~~~~~~~~~~~~~~~~~~~~

This module contains the main logic and highest level management objects along
with their configuration.

:copyright: (c) 2016 by Joe Black.
:license: Apache2.
"""

import time
import logging
import socket

from . import config, couch, kube, util
from .exceptions import InvalidKubeHostnameError

ONE_DAY = 60 * 60 * 24
log = logging.getLogger(__name__)


class ContainerEnvironment(util.ReprMixin):
    """Represents a self configuring environment object that can be passed to
    initialize other objects.
    """
    _public_attrs = ('index', 'petset', 'cluster_size', 'ports', 'creds')

    def __init__(self, env=None, host=None):
        self.env = env
        self._setup_environment(host)

    def _get_host(self, host=None):
        if not host:
            if self.env == 'dev':
                host = config.DEV_HOST
            else:
                host = socket.getfqdn()
        self._raise_if_host_invalid(host)
        return kube.KubeHostname(host)

    @staticmethod
    def _test_host(host):
        return len(host.split('.')) > 4

    def _raise_if_host_invalid(self, host):
        if not self._test_host(host):
            raise InvalidKubeHostnameError(host)

    def _setup_environment(self, host=None):
        self.host = self._get_host(host)
        self.kube = kube.KubeInterface(self.host, env=self.env)

    def reload(self):
        """Reload environment."""
        self._setup_environment()

    @property
    def index(self):
        """Represents the ordinal index 0-N of current node."""
        return self.host.index

    @property
    def petset(self):
        """Returns the petset name of the current node."""
        return self.host.petset

    @property
    def ports(self):
        """Returns the ports used by the CouchDB petset."""
        return self.kube.ports

    @property
    def creds(self):
        """Returns the auth credentials for the CouchDB petset."""
        return self.kube.creds

    @property
    def cluster_size(self):
        """Returns the expected cluster size of the CouchDB petset."""
        return self.kube.cluster_size

    @property
    def first_node(self):
        """Returns True if first node in cluster."""
        return self.index == 0

    @property
    def last_node(self):
        """Returns True if last node in cluster."""
        return (self.index + 1) == self.cluster_size

    @property
    def single_node_cluster(self):
        """Returns True if cluster is a single node cluster."""
        return self.cluster_size == 1


class ClusterManager(util.ReprMixin):
    """Represents the cluster manager, containing main logic that drives the
    self configuring of a CouchDB 2.0 cluster.
    """
    _public_attrs = ('env', 'couch')

    def __init__(self, env=None, host=None):
        self.env = ContainerEnvironment(env, host)
        self.couch = couch.CouchManager(self.env)

    def sleep_forever(self):
        """Work here is done, sleep forever.

        This would preferrably exit, but kubernetes only allows RestartPolicy
        to be applied per pod.
        """
        log.info('Done with: %s, sleeping forever', self.couch)
        while True:
            time.sleep(ONE_DAY)

    def run(self):
        """Main logic here, this is where we begin once all environment
        information has been retrieved."""
        log.info('Starting couchdiscover: %s', self.couch)
        if self.couch.disabled:
            log.info('Cluster disabled, enabling')
            self.couch.enable()
        elif self.couch.finished:
            log.info('Cluster already finished')
            self.sleep_forever()

        if self.env.first_node:
            log.info("Looks like I'm the first node")
            if self.env.single_node_cluster:
                log.info('Single node cluster detected')
                self.couch.finish()
        else:
            log.info("Looks like I'm not the first node")
            self.couch.add_to_master()
            if self.env.last_node:
                log.info("Looks like I'm the last node")
                self.couch.finish()
            else:
                log.info("Looks like I'm not the last node")
        self.sleep_forever()



"""
CouchDiscover
~~~~~~~~~~~~~

CouchDB 2.0 Autodiscovery for Kubernetes Environments.

Example:
    >>> import couchdiscover
    >>> couchdiscover.entrypoints.main()
"""

from . import config, util, exceptions, kube, couch, manage, entrypoints
from .kube import KubeHostname, KubeAPIClient, KubeInterface
from .couch import CouchServer, CouchInitClient, CouchManager
from .manage import ClusterManager, ContainerEnvironment
from .exceptions import (
    CouchDiscGeneralError,
    CouchDiscHTTPError,
    CouchAddNodeError,
    InvalidKubeHostnameError
)

__title__ = 'couchdiscover'
__version__ = '0.2.2'
__build__ = 0x000202
__author__ = "Joe Black <joe@valuphone.com>"
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2016 Joe Black'


util.setup_logging(
    level=config.LOG_LEVEL, fmt=config.LOG_FORMAT, date=config.DATE_FORMAT)



"""
couchdiscover.config
~~~~~~~~~~~~~~~~~~~~

This module holds configuration constants.

:copyright: (c) 2016 by Joe Black.
:license: Apache2.
"""

import os

ENVIRONMENT = os.getenv('ENVIRONMENT', 'production')

LOG_FORMAT = '%(asctime) %(levelname) %(module)s.%(funcName)s] %(message)s'
DATE_FORMAT = '%m/%d/%Y %I:%M:%S %p'
LOG_LEVEL = os.getenv('LOG_LEVEL', 'DEBUG').upper()

DEFAULT_CREDS = ('admin', 'secret')
DEFAULT_PORTS = (5984, 5986)

DEV_KUBECONFIG_PATH = "~/.kube/config"
DEV_HOST = 'couchdb-0.couchdb.default.svc.cluster.local'

del os



"""
couchdiscover.util
~~~~~~~~~~~~~~~~~~~

This module contains various utility functions and Mixins.

:copyright: (c) 2016 by Joe Black.
:license: Apache2.
"""

import logging


# needs to have a reference to the wrapped object @ self._wrapped
def passthrough(func):
    """A Decorator that will call the same named function on the wrapped
    object.

    Expected interface requires a `_wrapped` attribute on the wrapper object
    to contain a pointer to the wrapped object.
    """
    def wrap(self, *args, **kwargs):
        """This function allows arguments to be passed."""
        caller = func.__name__
        method = getattr(self._wrapped, caller)
        return method(*args, **kwargs)
    return wrap


def setup_logging(level, fmt, date):
    """This will setup logging for the module."""
    log = logging.getLogger('couchdiscover')
    loglvl = getattr(logging, level)
    log.setLevel(loglvl)

    chand = logging.StreamHandler()
    chand.setLevel(loglvl)
    chand.setFormatter(logging.Formatter(fmt, date))
    log.addHandler(chand)

    logging.raiseExceptions = False
    return log


class ReprMixin:
    """A mixin that does automatic repr configuration.

    Expected interface requires a `_public_attrs` attribute which is an
    iterable of public properties you want to expose through `__repr__`.
    """
    _public_attrs = ()

    def __repr__(self):
        clss = type(self).__name__
        attrs = ['{}: {}'.format(a, getattr(self, a))
                 for a in self._public_attrs]
        attrs = ', '.join(attrs)
        return '{}({})'.format(clss, attrs)



"""
couchdiscover.exceptions
~~~~~~~~~~~~~~~~~~~~~~~~

This module contains custom exceptions for use in this package, some which
add context.

:copyright: (c) 2016 by Joe Black.
:license: Apache2.
"""

import collections


class CustomErrorMixin:
    """Provides the ability to customize exceptions with context

    Usually works off of a message template stored in :attr:`_msg` by
    merging it as such self._msg.format(**kwargs).

    If args are provided, they are interpreted as usual and added to the end.

    The methods of this mixin are pretty generic and overriding them allows
    for much more advanced customization of the context.
    """

    _msg = ''

    def __init__(self, *args, **kwargs):
        kwargs = self._process_kwargs(kwargs)
        args = self._process_args(args, kwargs)
        super().__init__(*args)

    def _process_kwargs(self, kwargs=None):
        if not kwargs:
            kwargs = {}
        kwargs = collections.defaultdict(str, kwargs)
        for key, val in kwargs.items():
            setattr(self, key, val)
        return kwargs

    def _process_msg(self, kwargs):
        msg = ''
        if kwargs:
            msg = self._msg.format(**kwargs)
        return msg

    def _process_orig_msg(self, args):
        msg, *args = args
        return msg % tuple(args)

    def _merge_msgs(self, msg, orig):
        if msg:
            if orig:
                msg += ' ' + orig
        else:
            msg = orig
        return msg

    def _process_args(self, args, kwargs):
        msg = self._process_msg(kwargs)
        print('msg: ', msg)
        if args:
            orig = self._process_orig_msg(args)
            print('orig: ', orig)
            msg = self._merge_msgs(msg, orig)
            print('msg: ', msg)
        args = (msg,)
        print('args: ', args)
        return args


class CouchDiscGeneralError(CustomErrorMixin, Exception):
    """Generic catch all exception.

    Catch this exception in a except block to catch all errors related to this
    package.
    """


class CouchDiscHTTPError(CouchDiscGeneralError):
    """Generic exception during HTTP request

    Adds additional context from request and response to error message.

    Add the `req` kwarg when raising error to attach request context.

    * Overrides `_process_kwargs` to parse out request and response
    * Overrides `_process_msg` to add req and resp context to error message
    * Overrides `_merge_msgs` to allow the orig message to provide the message

    Example:
    >>> import requests
    >>> r = requests.get('http://httpbin.org/user-agent')
    >>> raise CouchDiscHTTPError('Error', req=r)
    CouchDiscHTTPError: Error req(GET http://httpbin.org/user-agent None)
    resp(200 {"user-agent": "python-requests/2.12.3"})
    """

    def _process_kwargs(self, kwargs=None):
        if not kwargs:
            kwargs = {}
        if kwargs.get('req'):
            kwargs['resp'] = kwargs.pop('req')
            if hasattr(kwargs.get('resp'), 'request'):
                kwargs['req'] = kwargs['resp'].request
        kwargs = super()._process_kwargs(kwargs)
        return kwargs

    def _merge_msgs(self, extra, orig):
        return super()._merge_msgs(orig, extra)

    def _process_msg(self, kwargs=None):
        if not kwargs:
            kwargs = {}
        extra = ''
        req = kwargs.get('req')
        if req:
            extra += 'req(%s %s %s)' % (req.method, req.url, req.body)
        resp = kwargs.get('resp')
        if resp:
            if req:
                extra += ' '
            extra += 'resp(%s %s)' % (resp.status_code, resp.text)
        return extra


class CouchAddNodeError(CouchDiscHTTPError):
    """Error adding node to master."""


class InvalidKubeHostnameError(CouchDiscGeneralError):
    """Invalid kubernetes hostname

    Raised when a host doesn't match the signature of a kubernetes petset
    hostname.

    Provides a `_msg` that includes the host passed in as a kwarg.

    Example:
    >>> raise InvalidKubeHostnameError(host='google.com')
    InvalidKubeHostnameError: The Hostname: google.com doesn't match the
    signature of a kubernetes petset hostname.
    """

    _msg = ("The Hostname: {host} doesn't match the signature of a kubernetes"
            " petset hostname.")



"""
couchdiscover.entrypoints
~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains functions that exist as entrypoints into the package.

:copyright: (c) 2016 by Joe Black.
:license: Apache2.
"""
import sys

from . import config, manage


def main(progname=sys.argv[0]):
    """main

    Main entrypoint executed by bin stub: `couchdiscover`.
    """
    man = manage.ClusterManager(env=config.ENVIRONMENT)
    return man.run()



import re
# import ez_setup
# ez_setup.use_setuptools()

from setuptools import setup

with open('couchdiscover/__init__.py', 'rt') as fd:
    VERSION = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not VERSION:
    raise RuntimeError('Cannot find version information')

with open('README.md', 'rt') as fd:
    README = fd.read()

setup(
    name='couchdiscover',
    version=VERSION,
    description='Autodiscovery & Clustering for CouchDB 2.0 with Kubernetes',
    long_description=README,
    keywords = ['couchdb', 'kubernetes', 'cluster'],
    author='Joe Black',
    author_email='joeblack949@gmail.com',
    url='https://github.com/joeblackwaslike/couchdiscover',
    download_url='https://github.com/joeblackwaslike/couchdiscover/tarball/0.2.2',
    license='Apache 2.0',
    packages=['couchdiscover'],
    package_data={'': ['LICENSE']},
    dependency_links=['https://github.com/kelproject/pykube/tarball/cd5b511474bf8854d122af371681fb1dc467a1b9#egg=pykube'],
    install_requires=['CouchDB', 'requests', 'pykube'],
    entry_points=dict(
        console_scripts=['couchdiscover = couchdiscover.entrypoints:main']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Database',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: System :: Clustering',
    ]
)



"""
couchdiscover.couch
~~~~~~~~~~~~~~~~~~~

This module contains constants and classes that abstract the details of
working with and clustering CouchDB 2.0 as higher level objects.

:copyright: (c) 2016 by Joe Black.
:license: Apache2.
"""

import time
import json
import logging
import socket

import requests
import couchdb

from . import config, util
from .exceptions import CouchAddNodeError


ADMIN_ONLY_DBS = ('_dbs', '_nodes', '_replicator', '_users')
log = logging.getLogger(__name__)


class CouchServer(util.ReprMixin):
    """Encapsulates the logic for interacting with CouchDB 2.0 Server"""
    _public_attrs = ('url', 'type', 'up')

    def __init__(self, proto='http', host='localhost',
                 port=config.DEFAULT_PORTS[0], creds=config.DEFAULT_CREDS):
        self._up = None
        self._args = dict(proto=proto, host=host, port=int(port), auth=creds)
        self.url = self._get_server_url()
        self._couch = couchdb.Server(self.url)
        self._wrapped = self._couch
        self._session = self._get_session()
        self.type = self._detect_type()

    def _get_server_url(self):
        args = self._args
        url = []
        url.append('{}://'.format(args['proto']))
        if args['auth']:
            url.append('{}:{}@'.format(*args['auth']))
        url.append(args['host'])
        url.append(':{}'.format(args['port']))
        url = ''.join(url)
        return url

    def _get_session(self):
        sess = requests.Session()
        sess.headers.update({'Content-Type': 'application/json'})
        if self._args['auth']:
            sess.auth = self._args['auth']
        return sess

    def _detect_type(self):
        all_dbs = self.all_dbs()
        if isinstance(all_dbs, list):
            if '_nodes' in all_dbs:
                return 'admin'
            else:
                return 'data'

    def _build_url(self, uri=None):
        if uri:
            if not uri.startswith('/'):
                uri = '/' + uri
        return self.url + uri

    @property
    def up(self):
        """Returns True if server is up."""
        try:
            if self.version():
                return True
        except ConnectionRefusedError:
            pass

    def __contains__(self, key):
        all_dbs = self.all_dbs()
        if all_dbs:
            return key in all_dbs

    def __getitem__(self, key):
        return self._couch[key]

    def __delitem__(self, key):
        del self._couch[key]

    @util.passthrough
    def config(self, *args, **kwargs):
        """Displays server configuration."""
        pass

    @util.passthrough
    def create(self, *args, **kwargs):
        """Create a database."""
        pass

    @util.passthrough
    def delete(self, *args, **kwargs):
        """Delete a database."""
        pass

    @util.passthrough
    def stats(self, *args, **kwargs):
        """Server stats."""
        pass

    @util.passthrough
    def version(self, *args, **kwargs):
        """Gets version of CouchDB."""
        pass

    def all_dbs(self):
        """Returns a generator iterating all DB objects."""
        return self.request(uri='/_all_dbs')

    def request(self, verb='get', uri=None, params=None, data=None,
                headers=None, files=None):
        """Send a low level HTTP request."""
        url = self._build_url(uri)
        sess = self._session
        try:
            req = sess.request(verb, url, params, data, headers, files=files)
            try:
                json_ = req.json()
                return json_
            except json.JSONDecodeError:
                return {}
        except requests.ConnectionError:
            pass


class CouchInitClient:
    """Encapsulates a pair of CouchServer objects for admin and data ports."""
    def __init__(self, host='localhost', ports=config.DEFAULT_PORTS,
                 creds=config.DEFAULT_CREDS, proto='http'):
        self._secure = False
        self._args = dict(
            proto=proto, host=str(host), ports=ports, creds=creds)
        self._wait_for_couch()
        self._servers = self._setup_servers()
        self._upgrade_auth_if_enabled()

    def _wait_for_couch(self):
        args = self._args
        url = 'http://{}:{}/_up'.format(args['host'], args['ports'][0])
        up = False
        log.info('Waiting for host: %s to be up', url)
        while not up:
            try:
                req = requests.get(url)
                up = True
                log.info('Host is up')
                del req
            except requests.RequestException:
                log.info('Host: %s not up yet, retrying in 5s', url)
                time.sleep(5)

    def _upgrade_auth_if_enabled(self):
        status = self.status
        if status == 'cluster_disabled':
            self._secure = False
        else:
            self._upgrade_auth()

    def _upgrade_auth(self):
        self._secure = True
        self._servers = self._setup_servers(auth=True)

    @property
    def status(self):
        """Returns the cluster_setup state string."""
        req = self.cluster_setup(action='status')
        if req.get('error'):
            state = 'auth_required'
        else:
            state = req['state']
        return state

    @property
    def disabled(self):
        """Returns True if status is `cluster_disabled`."""
        return self.status == 'cluster_disabled'

    @property
    def enabled(self):
        """Returns True if status is `cluster_enabled`"""
        return self.status == 'cluster_enabled'

    @property
    def finished(self):
        """Returns True if status is `cluster_finished`"""
        return self.status == 'cluster_finished'

    def __repr__(self):
        cls_name = self.__class__.__name__
        servers = ['{}: {}'.format(k, v) for k, v in self._servers.items()]
        servers = ', '.join(servers)
        attrs = 'status: {}, servers: {}'.format(self.status, servers)
        return '{cls}({attrs})'.format(cls=cls_name, attrs=attrs)

    def __str__(self):
        return self._args['host']

    def __getitem__(self, key):
        server = self._server_for(key)
        return server[key]

    def _setup_servers(self, auth=False):
        servers = {}
        args = self._args
        for port in args['ports']:
            this_auth = args['creds'] if auth else None
            server = CouchServer(args['proto'], args['host'], port, this_auth)
            servers[server.type] = server
        return servers

    def _server_for(self, db):
        if db in ADMIN_ONLY_DBS:
            key = 'admin'
        else:
            key = 'data'
        return self._servers[key]

    def call(self, server, method, *args, **kwargs):
        """Call a bound method of the CouchServer object."""
        server = self._servers.get(server)
        bound_method = getattr(server, method)
        return bound_method(*args, **kwargs)

    def request(self, server='data', verb='get', uri=None, params=None,
                data=None, headers=None, files=None):
        """Wraps the `CouchServer.request`, dispatching by `server`."""
        server = self._servers.get(server)
        return server.request(verb, uri, params, data, headers, files)

    def _build_cluster_setup_payload(
            self, action='add', host=None, port=None, creds=None):
        """Builds a data payload for `request`."""
        if not host:
            host = self._args['host']
        if not port:
            port = self._args['ports'][0]
        if not creds:
            creds = self._args['creds']

        data = None
        if action == 'add':
            data = {'action': 'add_node', 'host': host, 'port': port}
        elif action == 'enable':
            data = {'action': 'enable_cluster'}
        elif action == 'finish':
            data = {'action': 'finish_cluster'}

        if action in ('add', 'enable') and creds and len(creds) == 2:
            data['username'], data['password'] = creds

        if isinstance(data, dict):
            data = json.dumps(data)
        return data

    def cluster_setup(self, action='add', host=None, port=None, creds=None):
        """Creates and submits a request to the `cluster_setup` endpoint."""
        if action == 'status':
            verb = 'get'
        elif action in ('add', 'enable', 'finish'):
            verb = 'post'
        else:
            raise requests.HTTPError('%s is not a valid action', action)

        data = self._build_cluster_setup_payload(action, host, port, creds)
        req = self.request(
            server='data', verb=verb, uri='/_cluster_setup', data=data)
        return req

    @staticmethod
    def host_is_valid(host):
        """Returns true if the host passed can be resolved by DNS."""
        try:
            socket.gethostbyname(host)
            return True
        except socket.gaierror:
            return False

    def _test_node(self, node):
        """Runs several tests and returns True if node passes."""
        host = node._args['host']
        if all([self.host_is_valid(host), node.up()]):
            if not self._node_in_nodes('couchdb@{}'.format(host)):
                return True

    def enable(self):
        """Enables the current node."""
        req = self.cluster_setup(action='enable')
        if req.get('ok'):
            self._upgrade_auth()

    def add_node(self, remote):
        """Add's a new node to the current node."""
        if self._test_node(remote):
            args = remote._args
            req = self.cluster_setup(
                'add', args['host'], args['ports'][0], args['creds'])
            if req and isinstance(req, dict) and req.get('ok') is True:
                return req
            else:
                raise CouchAddNodeError(
                    'error adding node: %s', remote, req=req)

    def finish(self):
        """Finish the cluster."""
        return self.cluster_setup(action='finish')

    def nodes(self):
        """Get all nodes in the _nodes db of current node."""
        return tuple([n for n in self['_nodes']])

    def _node_in_nodes(self, node):
        """Return True if `node` is in _nodes db of current node."""
        nodes = self.nodes()
        return node in nodes

    def membership(self):
        """Returns the results of the `/_membership` endpoint."""
        return self.request(server='data', uri='/_membership')

    def up(self):
        """Returns if both CouchServer's return True for `s.up`"""
        return all([s.up for s in self._servers.values()])


class CouchManager:
    """Contains configuration data and topology of couch cluster."""
    def __init__(self, env):
        self.host = env.host
        self.ports = env.ports
        self.creds = env.creds
        self.local = CouchInitClient(env.host, env.ports, env.creds)
        if not self.is_master:
            mhost = env.host.clone(master=True)
            self.master = CouchInitClient(mhost, env.ports, env.creds)

    def __repr__(self):
        clss = type(self).__name__
        attrs = ['{}: {}'.format(a, getattr(self, a)) for a in
                 ('status', 'is_master', 'ports', 'creds')]
        attrs = ', '.join(attrs)
        return '{}({})'.format(clss, attrs)

    @property
    def status(self):
        """Returns the `status` of wrapped local node."""
        return self.local.status

    @property
    def enabled(self):
        """Returns the `enabled` of wrapped local node."""
        return self.local.enabled

    @property
    def disabled(self):
        """Returns the `disabled` of wrapped local node."""
        return self.local.disabled

    @property
    def finished(self):
        """Returns the `finished` of wrapped local node."""
        return self.local.finished

    @property
    def is_master(self):
        """Returns whether local node is the `master`."""
        return self.host.index == 0

    def enable(self):
        """Enable the local node but with error checking and logging."""
        if self.enabled:
            log.warning('Already enabled')
        elif self.finished:
            log.warning("Can't enable finished cluster: %s", self.local)
        else:
            log.info('Enabling local: %s', self.local)
            return self.local.enable()

    def finish(self):
        """Finish cluster but with error checking and logging."""
        if self.disabled:
            log.warning("Can't finish cluster when disabled: %s", self.local)
        elif self.finished:
            log.warning('Cluster already finished: %s', self.local)
        else:
            log.info('Finishing cluster: %s', self.local)
            return self.master.finish()

    def wait_for_enabled_master(self):
        """Blocking wait until master is enabled.

        Prevents a race condition where non-master node tries to
        manipulate the state of the cluster before the master is enabled.
        """
        if self.is_master:
            log.warning("Can't wait for master when master: %s", self.local)
        else:
            while not self.master.enabled:
                log.info('Waiting for master: %s to be enabled', self.master)
                time.sleep(5)

    def add_to_master(self, node=None):
        """Add the local node to master with error checking and logging."""
        if not node:
            node = self.local
        if self.is_master:
            log.warning("Can't add self to self, master: %s", self.master)
        else:
            self.wait_for_enabled_master()
            log.info('Adding: %s to master: %s', node, self.master)
            return self.master.add_node(node)
